
#define SECTOR_SIZE 512  // this code just can handle 512 bytes 
#define BULK_SIZE    64  // allways 64 byte on bulk for USB FS

#include "..\inc\mytypes.h"
#include "..\inc\wcherror.h"
#include "..\inc\fat32.h"

#include "..\inc\usb\usb11.h"
#include "..\inc\usb\usbmsd10.h"
#include "..\inc\ch559usb.h"
#include "..\inc\ch559_keil.h"

extern uint32_t SwapReadLong(uint8_t xdata *buffer,uint16_t Offset);
extern uint16_t SwapReadWord(uint8_t xdata *buffer,uint16_t Offset);
extern uint8_t  IAP_Prom(uint8_t xdata *buffer,uint16_t len,uint8_t SectorNum);
extern uint8_t  Host_UsbTransact(uint8_t endp_pid, uint8_t tog, uint16_t timeout);
extern uint8_t  Host_CtrlTransfer(uint8_t xdata *DataBuf, uint8_t idata *RetLen);

extern uint8_t xdata RxBuffer[];
extern uint8_t xdata TxBuffer[];
extern uint8_t code  Fname[];

extern uint8_t DevEndpOUTAddr;
extern uint8_t DevEndpINAddr;
extern uint8_t CurrentLun;

// fat32 variable
uint32_t data PartStart;     // sector where partition start
uint32_t data PartDataStart; // sector where data starts
uint32_t data Rootcluster;   // cluster for the root directory
uint32_t data FileStart;     // sector where the binary starts
uint32_t data ReservedSect;  // reserved by the system
uint16_t data FileSize;      // Filesize of hardcoded file notice 16 bit
uint8_t  data MaxCluster;


/* compare code with root filename in dir entry 
   return 0 if the name matches otherwise 1
*/
uint8_t strComp(uint8_t code *source, uint8_t xdata *target, uint8_t len)
{
   uint8_t i;
   for(i=0;i<len;i++)
   { 
      if(*source != *target)  return 1;
      source ++;
      target ++;	
   }
   return 0;
}

/* Read a secor using bulk only cmd
 
   This function is optimized. Since there is just a read cmd 
   the BulkOnly struct is setup directly. flag and remLen are
   const so they dont have to be checked.
   DataBUf:  ptr to a sectorbuffer filled with the sector
   SectorNo: absolute sector to be read 
   return    0 ir ome error code  
*/
uint8_t UDisk_ReadOneSector(uint8_t *DataBuf, uint32_t SectorNo)
{
   uint8_t  error;
   //CBW      cbw;
   uint16_t remLen;
   uint8_t  i;
   //
   // prepare the cmd this way because this is shorter than 
   // using some memcpy together with a rom const
   //
   TxBuffer[ 0] ='U'; // signature
   TxBuffer[ 1] ='S';
   TxBuffer[ 2] ='B';
   TxBuffer[ 3] ='C';
   
   TxBuffer[ 4] = 0;  // tag ToDo what about something different 
   TxBuffer[ 5] = 0;
   TxBuffer[ 6] = 1;
   TxBuffer[ 7] = 0; 
   
   TxBuffer[ 8] = SECTOR_SIZE & 0xFF;
   TxBuffer[ 9] = SECTOR_SIZE >> 8;
   TxBuffer[10] = 0;
   TxBuffer[11] = 0;
   TxBuffer[12] = 0x80;        // flag reading
   TxBuffer[13] = CurrentLun;  // lun always 0
   TxBuffer[14] = 10;          // cblen 10bytes
   TxBuffer[15] = SPC_CMD_READ10;
   TxBuffer[16] = CurrentLun << 4;
#ifdef __C51__
   (uint32_t) TxBuffer[17] = SectorNo;
#else
   TxBuffer[17] = SectorNo >> 24;  //len msbfirst
   TxBuffer[18] = SectorNo >> 16;
   TxBuffer[19] = SectorNo >>  8;
   TxBuffer[20] = SectorNo >>  0;
#endif
   TxBuffer[21] = 0;     // expected tag
   TxBuffer[22] = 0;
   TxBuffer[23] = 1;
   TxBuffer[24] = 0;

   TxBuffer[25] = 0;     // clear the rest
   TxBuffer[26] = 0;
   TxBuffer[27] = 0;
   TxBuffer[28] = 0;
   TxBuffer[29] = 0;
   TxBuffer[30] = 0;
  
   //
   // transmit the BulkOnlyCmd
   //
	 UH_TX_LEN   = 31;
	 UH_TX_CTRL ^= bUH_T_TOG;
	 error = Host_UsbTransact(USB_PID_OUT << 4 | DevEndpOUTAddr, 
	                          UH_TX_CTRL, 
	                          10000);
   if (error) return error;
   //
   // read the sector using  64 byte junks
   //
   remLen = SECTOR_SIZE;
   while(remLen)
   {
      error = Host_UsbTransact(USB_PID_IN << 4 | DevEndpINAddr, 
                               UH_RX_CTRL, 
                               65535);
      if (error) return error;

      for (i=0 ;i < BULK_SIZE; i++)
      {
      	 *DataBuf = RxBuffer[i];
      	 DataBuf++;           	  
      }
      remLen  -= BULK_SIZE;
   }
   error = Host_UsbTransact(USB_PID_IN << 4 | DevEndpINAddr, 
                            UH_RX_CTRL, 
                            10000); 
   if (error) return error;
   return RxBuffer[0x0c]; //status byte
}

/* read the bootsector 
   The function first read the MBR / sector 0 to get the partitiontable.
   There it looks for the first active partition and if found reads the 
   boot record of that partition. If all is ok some later needed vars are
   set. 
*/
uint8_t UDisk_ReadDiskDBR(uint8_t xdata *DataBuf)
{
   uint8_t  error;
   uint8_t  partitions;
   uint16_t offset;
   uint32_t sector;
   //
   // read the Masterbootrecord / sector 0
   //
   error = UDisk_ReadOneSector(DataBuf, 0L);
   if(error) return error;
   if (SwapReadWord(DataBuf,wSignature)!=0xAA55) return ERR_MBR_ERROR;
   //
   // find the partitiontable and read the bootrecord
   //
   offset = PartTableStart;
   partitions = 4;
   while(partitions)
   {
      sector = SwapReadLong(DataBuf,offset+lFirstSector); //Part Start
      if (DataBuf[offset+bActiveFlag ]==0x80)
      {
         error = UDisk_ReadOneSector(DataBuf, sector);
         if (error) return error;
         break;
      }
      partitions --;
      offset +=16; // partiion entrie are 16 bytes long
   }
   if(partitions == 0) return ERR_PART_ERROR; //no active partition found
   //
   // here the first active bootsector is loaded
   //
   if (SwapReadWord(DataBuf,wSignature)!=0xAA55)             return ERR_FAT_ERROR;
   if (SwapReadWord(DataBuf,wBytesPerSector) != SECTOR_SIZE) return ERR_FAT_ERROR;

   if( (DataBuf[0]==0xEB) &&  // JMP
       (DataBuf[1]==0x58) &&  // Marker FAT32
       (DataBuf[2]==0x90))    // JMP offset
   {
      PartStart = sector;
      sector   +=  SwapReadWord(DataBuf,wReservedSectors);
      sector   += (SwapReadLong(DataBuf,lSectorsPerFat) * DataBuf[bNumberOfFats]);
      PartDataStart = sector;
      Rootcluster   = SwapReadLong(DataBuf,lRootCluster);
      MaxCluster    = DataBuf[bSectorsPerCLuster];
      ReservedSect  = SwapReadWord(DataBuf,wReservedSectors);
      return ERR_SUCCESS;
   }
   return ERR_FAT_ERROR; // not a FAT32 partiton
}

/* get a new cluster from the FAT for file read */
uint32_t UDisk_ReadDiskFAT(uint8_t xdata *DataBuf, uint32_t ClusNum)
{
   uint8_t  error;
   uint16_t j;
   uint32_t sectorNumber;
   uint32_t nextClusNum;
   //
   // calculate the sector where the FAT starts
   //
   //sectorNumber  = (ClusNum << 2) / SECTOR_SIZE; //original
   sectorNumber  = ClusNum >> 7;                   // optimized
   sectorNumber += PartStart;
   sectorNumber += ReservedSect;

   error = UDisk_ReadOneSector(DataBuf,sectorNumber);
   if(error == ERR_SUCCESS)
   {
      j = (ClusNum*4)%(SECTOR_SIZE);  // orginal todo is there a better way?
      nextClusNum = SwapReadLong(DataBuf,j);
   }
   if( nextClusNum == 0x0fffffff) return 0; //no more file clusters
   return nextClusNum;
}

/* serch for the hard coded filename in the rootdir 
   look in the root dir for the specified file. If found check the filesize
   for less than 64k because max file cant be bigger  
*/
uint8_t UDisk_FileOpen(uint8_t xdata *DataBuf)
{
   uint8_t   error;
   uint8_t   j;
   uint16_t  i;
   uint32_t  nextClusNum;
   uint32_t  sectnum;

   nextClusNum = Rootcluster;
   do
   {
      sectnum  = MaxCluster*(nextClusNum-Rootcluster);
      sectnum += PartDataStart;
      for(j=0; j < MaxCluster; j++)
      {
         //printf("FileOpen:ClusterSect: 0x%8.8X\n",sectnum+j);
         error = UDisk_ReadOneSector(DataBuf,sectnum+j);
         if(error == 0/* ERR_SUCCESS*/)
         {
            //PrintSector(Buffer,32);
            //printf("\n\n");
            i = 0; //find the filename
            while (strComp(Fname, &DataBuf[i],11) && (i<512))
            {
               i += DIR_ENTRY_SIZE;
            }

            if(i < SECTOR_SIZE)
            {
               //just allow files < 64k
               if (SwapReadWord(DataBuf,i+lFileSize+2) !=0) return ERR_TOO_LARGE;
               FileSize  = SwapReadWord(DataBuf,i+lFileSize);
               // a hack to avoid 32 bit shift which is costly on c51
               FileStart = SwapReadLong(DataBuf,i+wFirstClusterHi-2);
               FileStart &=0xFFFF0000;      // trow the lowpart
               FileStart |=SwapReadWord(DataBuf,i+wFirstClusterLo);
               return ERR_SUCCESS;
            }
            else  error = ERR_MISS_FILE;
         }
      }
      nextClusNum = UDisk_ReadDiskFAT(DataBuf,nextClusNum);
   } while (nextClusNum);
   return error;
}

/* read the file and flash it starting at address 0x0000.
   The code autamatically deletes the flashpages ocupied by the file
   and flashs the filecontents
*/
uint8_t UDisk_ReadFile(uint8_t xdata *DataBuf)
{
   uint8_t   error, j;
   uint8_t   filesec;
   uint16_t  len;
   uint32_t  nextClusNum;
   uint32_t  secnum;
   uint32_t  resvsize;

   resvsize    = FileSize;
   nextClusNum = FileStart;
   filesec = 0;
   do
   {
      secnum  = MaxCluster*(nextClusNum-Rootcluster);
      secnum += PartDataStart;

      for(j = 0; j < MaxCluster; j++)
      {
         error = UDisk_ReadOneSector(DataBuf,secnum+j);
         if(error) return error;
         len = (resvsize > SECTOR_SIZE) ? SECTOR_SIZE
                                        : resvsize;
         IAP_Prom(DataBuf, len, filesec);
         filesec ++;
         resvsize -= len;
         if(!resvsize) return error;
      }
      nextClusNum = UDisk_ReadDiskFAT(DataBuf,nextClusNum);
   } while(nextClusNum);
   return error;
}
